/*============================================================================
 * filename: ../dss/agent/dslib.c
 *
 * $Id: dslib.c,v 1.5 1996/08/27 15:13:00 root Exp $
 * copyright 1996 (c) Digital Pathways, Inc.  All Rights Reserved
 *
 * This source code is provided under license from Digital Pathways,
 * Inc.  It is provided "AS-IS", without warranty of any kind, either
 * express or implied. It may be incorporated wholly or modified as
 * needed, subject to the following restrictions.
 *
 * 1) This notice, including the Digital Pathways, Inc. notice of
 * copyright must be included in any source code or derivative works.
 * If this source code is modified by the licensee, a notice of
 * modification must be included as well.
 *
 * 2) This Program, and any derivative works, may not be distributed
 * is source form, on any media, without the express written consent
 * of Digital Pathways, Inc.  There is no restriction on object or
 * executable code.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>  
#ifndef __hppa
#include 	<sys/select.h>		/* to support fd_set 		  */
#endif 
#ifdef _WIN32
 #include <windows.h>
 #include <winsock.h>
 #include <memory.h>
#else
 #include <sys/socket.h>
 #include <netdb.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <unistd.h>
 #include <sys/time.h>
#endif
#include <time.h>
#include <stdarg.h>	
#include "dslib.h"
#include "dsagent.h"

static char rcsid[] = "$Id: dslib.c,v 1.5 1996/08/27 15:13:00 root Exp $";

static cfg_t *agent_cfg;
int da_errno;

#ifdef _WIN32
/*
 * 
 * This is the entry point for the library if you make a Win32 dll.
 * This doesn't actually do anything, it is here to make the import
 * library for the dll.  You may wish to initialize the winsock layer
 * in DLL_PROCESS_ATTACH and make sure that all sockets are closed in
 * DLL_PROCESS_DETACH.
 */
 
BOOL WINAPI
DllMain(HANDLE hinstDLL, ULONG Reason, LPVOID lpvReserved)
{
	switch (Reason) {
	case DLL_PROCESS_ATTACH:
		break;

	case DLL_THREAD_ATTACH:
	case DLL_THREAD_DETACH:
		break;

	case DLL_PROCESS_DETACH:
		break;
	}
	return(TRUE);
}
#endif

/*
 * GetRandomLong()
 * 
 * Return a reasonably random long.  Since this is only used
 * for the initial challenge to the DSS, it is sufficient.
 */
static unsigned long
GetRandomLong(void)
{
#ifdef _WIN32
	static DWORD lastRandom;

	return(lastRandom ^= GetTickCount());
#else
	struct timeval t;
	struct timezone tz;

	gettimeofday(&t, &tz);
	return(t.tv_usec);
#endif
}

/*
 * snk_response
 * Compute a Binary SNK Response
 *
 * Given a pre-permuted key array, and pointers to a long challenge and
 * a location for a long response, compute the appropriate response
 * to the challenge.  Please note that this is a one way function; 
 * although it uses DES to generate the response, half the result is
 * discarded.  The response cannot be used to generate the challenge.
 */

static unsigned long
snk_response(unsigned long challenge)
{
	unsigned char data[8];
	unsigned long response;

	/*
	 * This translation enforces little endian byte ordering.  Although
	 * elsewhere we assume big endian byte ordering, we do this here
	 * for historical (hysterical?!?) reasons.  Please note that the 
	 * original code gets around this not by a change in this function
	 * on big endian machines, but by not swapping the challenge and
	 * response fields.
	 */
	data[0] = (unsigned char)(challenge & 0xff);
	data[1] = (unsigned char)(challenge >> 8 & 0xff);
	data[2] = (unsigned char)(challenge >> 16 & 0xff);
	data[3] = (unsigned char)(challenge >> 24 & 0xff);
	data[4] = 0;
	data[5] = 0;
	data[6] = 0;
	data[7] = 0;
	da_des_encrypt((char *)data, (char *)data, agent_cfg->da_cfg.pkeys);

	/*
	 * similarly, little endian byte ordering is enforced.
	 */
	response = data[3] << 24 | data[2] << 16 | data[1] << 8 | data[0];
	return(response);
}

/*
 * next_session_id()
 *
 * get next session ID.  0 and negative numbers are special and
 * not used
 */

static short
next_session_id(void)
{
	static short next_id;

	if (++next_id <= 0)
		next_id = 1;
	return (next_id);
}

/*
 * get_session_by_id()
 *
 * Find the session matching the session id.
 */
static session_t *
get_session_by_id(int id)
{
	session_t *s;

	for (s = agent_cfg->sessions; s && s->session_id != id; s = s->next);
	return(s);
}

/*
 * get_session_by_tid()
 *
 * Find the session matching the transaction id.
 */
static session_t *
get_session_by_tid(long tid)
{
	session_t *s;

	for (s = agent_cfg->sessions; s && s->transaction_id != tid; s = s->next);
	return(s);
}

/*
 * free_dialog()
 *
 * free the dialog structure.  There is no reference or consistency
 * checking done here.
 */
static void
free_dialog(dialog_t *dialog)
{
	dialog->input = NULL;
	dialog->prompt = NULL;
	free(dialog);
}

/*
 * alloc_dialog()
 *
 * Allocate a dialog record.  The prompt and input pointers are set
 * up here so that those two fields in the packet can be easily 
 * referenced.
 */
static dialog_t *
alloc_dialog(void)
{
	dialog_t *dialog;

	dialog = malloc(sizeof(dialog_t));
	if (dialog) {
		/*
		 * Initialize the packet to null except for the link field, which
		 * is set the the identifier value of 2.
		 */
		memset(dialog, 0, sizeof(dialog_t));
		dialog->packet[LINK_FIELD] = LINK_VALUE;
		/*
		 * set up the conveinience pointers for the prompt and input fields
		 */
		dialog->prompt = &dialog->packet[PROMPT_FIELD];
		dialog->input = &dialog->packet[INPUT_FIELD];
	}
	else
		da_errno = DA_ALLOC_FAILED;
	return(dialog);
}

/*
 * free_session()
 *
 * release the session structure.  Make sure it it off the active
 * session list, and release any associated dialogs. 
 */
static void
free_session(session_t *session)
{
	session_t **s;

	for (s = &agent_cfg->sessions; *s && *s != session; s = &(*s)->next);
	if (*s) {
		*s = (*s)->next;
		agent_cfg->active_sessions--;
		free(session);
	}
}

/*
 * alloc_session()
 *
 * Allocate a session structure and assign it a session id.
 * We ensure that the session is on the active session list mostly for
 * tracking.
 */
static session_t *
alloc_session(void)
{
	session_t *session;

	/*
	 * Check to see if we have exceeded the maximum session count.  Note
	 * that if max_sessions == 0, then the count of active sessions is
	 * unlimited.
	 */
	if (agent_cfg->da_cfg.max_sessions &&
		agent_cfg->da_cfg.max_sessions <= agent_cfg->active_sessions) {
		da_errno = DA_SESSION_LIMIT;
		return(NULL);
	}
		
	session = malloc(sizeof(session_t));
	if (session) {
		memset(session, 0, sizeof(session_t));
		/*
		 * initialize the callslot and transaction id to values the 
		 * server recognizes as unassigned.
		 */
		session->callslot = -1;
		session->transaction_id = -1L;
		session->session_id = next_session_id();
		session->in_addr = ((struct sockaddr_in *)&agent_cfg->sa)->sin_addr;
		session->next = agent_cfg->sessions;
		agent_cfg->sessions = session;
		agent_cfg->active_sessions++;
	}
	else
		da_errno = DA_ALLOC_FAILED;
	return(session);
}

/*
 *	write_info()
 *
 *	Given a pointer to a dialog structure, fills
 *	the event structure with the appropriate information
 */

static void
write_info(dialog_t *dialog)
{
	char *buf;

	buf = dialog->prompt;
	*buf++ = agent_cfg->da_cfg.max_sessions & 0xff;
	*buf++ = agent_cfg->da_cfg.max_sessions >> 8 & 0xff;
	if (agent_cfg->da_cfg.flags & REPORT_END_OF_CALL) {
		*buf++ = 1;
		*buf++ = 0;
	}
	else {
		*buf++ = 0;
		*buf++ = 0;
	}
	buf += 2;
	if (agent_cfg->da_cfg.flags & PRIMARY_SERVER) {
		*buf++ = 1;
		*buf++ = 0;
	}
	else {
		*buf++ = 0;
		*buf++ = 0;
	}
	strncpy(&dialog->prompt[IF_AGENTDESC], AGENT_DESC, AGENT_DESC_LEN);
	strncpy(&dialog->prompt[IF_AGENTVERS], AGENT_VER, AGENT_VER_LEN);
}

/*
 *	read_auth()
 *
 *	Given a pointer to a dialog structure, fills
 *	the event structure with the appropriate information
 */

static void
read_auth(dialog_t *dialog, auth_t *auth)
{
	unsigned char *buf;

	/*
	 * Note: we use an unsigned char * in order to avoid problems
	 * with sign extension during the conversion from canonical
	 * form.
	 */
	buf = (unsigned char *)&dialog->input[DP_AF_STATE];
	auth->state = buf[1] << 8 | buf[0];

	buf += DP_AF_CHALLENGE - DP_AF_STATE;
	auth->challenge = buf[3] << 24 | buf[2] << 16 | buf[1] << 8 | buf[0];
	buf += DP_AF_RESPONSE - DP_AF_CHALLENGE;
	auth->response = buf[3] << 24 | buf[2] << 16 | buf[1] << 8 | buf[0];

	buf += DP_AF_KEY - DP_AF_RESPONSE;
	auth->session_key[0] = buf[3] << 24 | buf[2] << 16 | buf[1] << 8 | buf[0];
	buf += 4;
	auth->session_key[1] = buf[3] << 24 | buf[2] << 16 | buf[1] << 8 | buf[0];
}

/*
 *	write_auth()
 *
 *	Given a pointer to a dialog structure, fills
 *	the event structure with the appropriate information
 */

static void
write_auth(dialog_t *dialog, auth_t *auth, char *agent_id)
{
	char *buf;

	buf = &dialog->input[DP_AF_STATE];
	*buf++ = auth->state & 0xff;
	*buf = auth->state >> 8 & 0xff;

	buf = &dialog->input[DP_AF_LEVEL];
	*buf++ = auth->level & 0xff;
	*buf = auth->level >> 8 & 0xff;

	buf = &dialog->input[DP_AF_CHALLENGE];
	*buf++ = (char)(auth->challenge & 0xff);
	*buf++ = (char)(auth->challenge >> 8 & 0xff);
	*buf++ = (char)(auth->challenge >> 16 & 0xff);
	*buf = (char)(auth->challenge >> 24 & 0xff);

	buf = &dialog->input[DP_AF_RESPONSE];
	*buf++ = (char)(auth->response & 0xff);
	*buf++ = (char)(auth->response >> 8 & 0xff);
	*buf++ = (char)(auth->response >> 16 & 0xff);
	*buf = (char)(auth->response >> 24 & 0xff);

	buf = &dialog->input[DP_AF_AGENTID];
	strncpy(buf, agent_id, AGENT_ID_LEN);
}

/*
 * process_dialog()
 *
 * allocate a dialog, read the packet from the socket and
 * parse the buffer from the byte packed, little-endian
 * canonical form to the native structure format.
 *
 * Note: we use an unsigned char * as the pointer into the 
 * packet buffer in order to avoid problems with sign extension
 * during the conversion from canonical form.
 */

static dialog_t *
process_dialog(void)
{
	dialog_t *dialog;
	unsigned char *buf;
	int cnt;
#ifdef _WIN32
	BOOL		result;
	DWORD		dwBytesRead;
	OVERLAPPED	OverLapped;

	ZeroMemory( &OverLapped, sizeof( OVERLAPPED));
#endif

	dialog = alloc_dialog();
	if (dialog) {
		/*
		 * Read the packet into the dialog packet buffer and error
		 * out if the count is not correct.  We do not attempt 
		 * accumulated reads because the packet size is < MSS and 
		 * the server ALWAYS writes the packet as a block.
		 */
#ifdef _WIN32
		result = ReadFile((HANDLE)agent_cfg->server_fd, dialog->packet,
			PACKET_SIZE, &dwBytesRead, &OverLapped);
		cnt = (int)dwBytesRead;
#else
		cnt = read(agent_cfg->server_fd, dialog->packet, PACKET_SIZE);
#endif
		if (cnt != PACKET_SIZE) {
			if (cnt < 0)
				da_errno = DA_RECEIVE_ERR;
			free_dialog(dialog);
			return(NULL);
		}

		/*
		 * Get the Call Slot number
		 */
		buf = (unsigned char *)dialog->packet;
		dialog->callslot = *buf | buf[1] << 8;
		/*
		 * Get my session ID value.  
		 */
		buf += SESSION_ID;
		dialog->session_id = *buf | buf[1] << 8;
		/*
		 * Get the Transaction ID assigned by the Server.
		 */
		buf += TID - SESSION_ID;
		dialog->tid = *buf | buf[1] << 8 | buf[2] << 16 | buf[3] << 24;
		/*
		 * Get the Command from the Agent to the Server.
		 */
		buf += AGENT_CMD - TID;
		dialog->cmd_from_agent = *buf | buf[1] << 8;
		/*
		 * Get the Command from the Server to the Agent.
		 */
		buf += DSS_CMD - AGENT_CMD;
		dialog->cmd_from_server = *buf | buf[1] << 8;
		/*
		 * Get the Response modifier from the Server to the Agent.
		 */
		buf += DSS_RESPONSE - DSS_CMD;
		dialog->server_response = *buf | buf[1] << 8;
		/*
		 * Get the line control value.
		 */
		buf += LINE_CTRL - DSS_RESPONSE;
		dialog->line_flags = *buf | buf[1] << 8;
		/*
		 * Get the dialog complete flag
		 */
		buf += CALL_COMPLETE - LINE_CTRL;
		dialog->call_complete = *buf | buf[1] << 8;
	}
	return(dialog);
}

/*
 * send_dialog()
 *
 * Put the dialog data into the transmission buffer in canonical form 
 * and write it out.  Note that although an error condition is noted,
 * it is merely passed back to the caller.  No action is taken at this
 * level.
 */

static int
send_dialog(int server_fd, dialog_t *dialog)
{
	char *buf;
	int cnt;
#ifdef _WIN32
	BOOL		result;
	DWORD		dwBytesWritten;
	OVERLAPPED	OverLapped;

	ZeroMemory( &OverLapped, sizeof( OVERLAPPED));
#endif

	buf = dialog->packet;
	*buf++ = dialog->callslot & 0xff;
	*buf++ = dialog->callslot >> 8 & 0xff;
	*buf++ = dialog->session_id & 0xff;
	*buf++ = dialog->session_id >> 8 & 0xff;
	*buf++ = dialog->tid & 0xff;
	*buf++ = dialog->tid >> 8 & 0xff;
	*buf++ = dialog->tid >> 16 & 0xff;
	*buf++ = dialog->tid >> 24 & 0xff;
	*buf++ = dialog->cmd_from_agent & 0xff;
	*buf++ = dialog->cmd_from_agent >> 8 & 0xff;
	/*
	 * We have to return the Server command and response fields
	 * because this dialog packet is used to maintain state and the
	 * interpretation of the agent command is dependent on these 
	 * fields.
	 */
	*buf++ = dialog->cmd_from_server & 0xff;
	*buf++ = dialog->cmd_from_server >> 8 & 0xff;
	*buf++ = dialog->server_response & 0xff;
	*buf++ = dialog->server_response >> 8 & 0xff;

#ifdef _WIN32
	result = WriteFile((HANDLE)server_fd, dialog->packet, PACKET_SIZE,
				&dwBytesWritten, &OverLapped);
	cnt = dwBytesWritten;			
#else
	cnt = write(server_fd, dialog->packet, PACKET_SIZE);
#endif

	free_dialog(dialog);
	if (cnt != PACKET_SIZE) {
		da_errno = DA_TRANSMISSION_ERR;
		return (ERROR);
	}
	else 
		return (SUCCESS);
}

/*
 * da_wait_server()
 *
 * wait up to the time limit specified in the configuration block
 * for a response from the server.  This is used by da_connect() and
 * da_authenticate(), since by definition those functions are single
 * threaded.  It is also made available to the host program, which may
 * optionally pass in a fd_set.
 */

int
da_wait_server(fd_set *rdfs)
{
	struct timeval timeout, *t;
	fd_set read_fds, *r;
	int rc;

	da_errno = SUCCESS;
	if (!rdfs) {
		FD_ZERO(&read_fds);
		r = &read_fds;
	}
	else
		r = rdfs;

	FD_SET(agent_cfg->server_fd, r);
	if (agent_cfg->da_cfg.timeout) {
		timeout.tv_sec = agent_cfg->da_cfg.timeout;
		timeout.tv_usec = 0;
		t = &timeout;
	}
	else
		t = NULL;
#ifdef __hppa
	rc = select(FD_SETSIZE, (int *)&read_fds, NULL, NULL, t);
#else	
	rc = select(FD_SETSIZE, &read_fds, NULL, NULL, t);
#endif	
	if (rc <= 0) {
		/*
		 * Timeout - return an error indication and set da_errno, since
		 * there is no error in errno.
		 */
		if (rc == 0)
			da_errno = DA_SERVER_TIMEOUT;
		else
			/*
			 * We let the host deal with the generalized error case
			 */
			da_errno = DA_RECEIVE_ERR;
	}
	return(rc);
}

/*
 * da_user_input()
 *
 * return a string input from the user to the DSS.
 *
 * Note:
 * The void *param passed in should be the same dialog structure passed
 * the the write function called in the callback from da_process_packet().
 * User input should be ignored until a dialog on a session id is posted
 * and once the user input is terminated, this function should be called
 * and the dialog should be cleared.  An error on user input, or timeout,
 * may be indicated by a -1 in the len field.  If the host chooses to 
 * have knowledge of the dialog structure and to use the dialog->input as
 * the user input buffer (watch out for overruns!), then that may be 
 * indicated by passing the buf = dialog->input.  The memcpy() will be 
 * skipped.
 *
 * Note: this function used to return an error status indication.  It is
 * now a void.  If the connection or session closes while waiting for
 * user input, this is the only way for the host to free the waiting dialog.
 */

int
da_user_input(char *buf, int len)
{
	session_t *session;
	dialog_t *dialog;
	int rc = SUCCESS;

	/*
	 * Check for a NULL passed in as a parameter
	 */
	if (!buf) {
		da_errno = DA_INVALID_PARAMETER;
		return(ERROR);
	}

	/*
	 * 'buf' is a pointer to the input field of the dialog and 
	 * itod simply subtracts the offset to the input field in 
	 * the packet to get back to the dialog.  MAKE SURE THAT WE
	 * THINK THE DIALOG IS VALID.
	 */
	dialog = itod(buf);
	if (dialog->input != buf) {
		da_errno = DA_INVALID_PARAMETER;
		return(ERROR);
	}

	da_errno = SUCCESS;
	/*
	 * If the length < 0, it indicates an error condition occurred during
	 * in the host function accepting user input.  We don't know what this
	 * is and we don't care.  It is up to the host to signal this.  We 
	 * must, however, release the dialog structure.
	 */
	if (len < 0) {
		free_dialog(dialog);
		return(SUCCESS);
	}

	/*
	 * make sure that the agent connection still exists.  If an agent
	 * connection closed while waiting for user input, this is the 
	 * only way to recycle the dialog.  This is not an error.
	 */
	if (agent_cfg && agent_cfg->server_fd != -1) {
		/*
		 * Find the session.  It's possible that the session no longer
		 * exists, so if it doesn't, just free the dialog and exit.
		 */
		if ((session = get_session_by_id(dialog->session_id))) {
			/*
			 * Make sure that the input field in the packet is not overrun.
			 */
			if (len > INPUT_LEN)
				len = INPUT_LEN;

			/*
			 * If we are in the "Ask for user id" phase of the dialog,
			 * copy the user id to the user id buffer.  The copy to the
			 * user supplied buffer will be done at the completion of the
			 * transaction.
			 */
			if (dialog->server_response == ASK_FOR_ID)
				strncpy(session->userid, buf, USER_ID_LEN);

			/*
			 * set up the dialog agent command field and send the packet.
			 */
			dialog->cmd_from_agent = USER_INPUT;
			if (send_dialog(agent_cfg->server_fd, dialog)) {
				if (session->callback.error)
					rc = (*session->callback.error)(session->session_id,
						da_errno);
				else
					rc = ERROR;
			}
		}
		else
			free_dialog(dialog);
	}
	else
		free_dialog(dialog);

	return(rc);
}

/*
 * da_process_packet()
 *
 * read, parse, and process the packet coming in from the DSS.
 */
int
da_process_packet()
{
	session_t *session;
	dialog_t *dialog;
	int rc = SUCCESS;

	da_errno = SUCCESS;
	/*
	 * Sanity check to make sure that the server is actually connected
	 * We set da_errno, but we do not call out to the error callback.
	 */
	if (!agent_cfg || (agent_cfg->server_fd == -1)) {
		da_errno = DA_NO_CONFIGURE;
		return(ERROR);
	}
	
	/*
	 * Read the packet and parse it out.
	 * da_errno is set in alloc_dialog();
	 */
	dialog = process_dialog();
	if (!dialog)
		return(ERROR);

	if ((session = get_session_by_id(dialog->session_id))) {
		/*
		 * Update the server callslot and transaction id fields
		 * in the session block.
		 */
		session->callslot = dialog->callslot;
		session->transaction_id = dialog->tid;

		switch (dialog->cmd_from_server) {
		case GET_INPUT:
			rc = (*session->callback.write)(dialog->session_id, 
				dialog->prompt, strlen(dialog->prompt));
			if (rc < 0) {
				rc = ERROR;
				da_errno = DA_LOST_USER_CONNECTION;
			}
			else {
				/*
				 * If the write succeeded, then the dialog is posted and we
				 * want to make sure that we don't free it unless an error is
				 * reported.
				 */
				rc = (*session->callback.read)(dialog->session_id,
					dialog->input, INPUT_LEN, dialog->server_response,
					dialog->line_flags);
				if (rc < 0)
					break;
				dialog = NULL;
				rc = 0;
			}
			break;

		case ACCESS_DENIED:
			/*
			 * Access denied - write out the prompt string and notify
			 * the host.
			 */
			rc = (*session->callback.write)(dialog->session_id,
				dialog->prompt, strlen(dialog->prompt));
			if (rc > 0)
				rc = (*session->callback.notify)(dialog->session_id,
						dialog->server_response, -1L, dialog->input);
			else if (rc == 0) {
				/*
				 * if rc == 0, it means that the user connection closed.
				 */
				rc = ERROR;
				da_errno = DA_LOST_USER_CONNECTION;
			}
			break;

		case ACCESS_APPROVED:
			/*
			 * user record, returned on successful ACCESS_APPROVED in the input
			 * field.
			 *
			 * offset 0  = char userID[10]
			 * offset 10 = char info1[10]
			 * offset 20 = char info2[10]
			 * offset 30 = char phone1[20]
			 * offset 50 = phone2[20]
			 * offset 70 = 0x00 terminator
			 */
			rc = (*session->callback.write)(dialog->session_id,
				dialog->prompt, strlen(dialog->prompt));
			if (rc > 0)
				rc = (*session->callback.notify)(dialog->session_id, SUCCESS,
					dialog->tid, dialog->input);
			else if (rc == 0) {
				/*
				 * if rc == 0, it means that the user connection closed.
				 */
				rc = ERROR;
				da_errno = DA_LOST_USER_CONNECTION;
			}
			break;

		case NO_COMMAND:
			/*
			 * We have two cases were the access is set.  It is possible
			 * to get a command granting or denying access, or there is no
			 * command and the response code carries the status of the access
			 * request.  This is pretty nonsensical, but must be dealt with.
			 * We only allow a limited set of the defined responses to occur
			 * here.
			 */
			/*
			 * write out the prompt string
			 */
			rc = (*session->callback.write)(dialog->session_id,
				dialog->prompt, strlen(dialog->prompt));
			if (rc > 0) {
				if (dialog->server_response == ACCESS_APPROVED) {
					rc = (*session->callback.notify)(dialog->session_id,
						SUCCESS, dialog->tid, dialog->input);
				}
				else if (dialog->server_response == ACCESS_DENIED) {
					/*
					 * On a denial, set the transaction ID = -1;
					 */
					rc = (*session->callback.notify)(dialog->session_id,
						dialog->server_response, -1L, dialog->input);
				}
				else {
					/*
					 * This is the same as ACCESS_DENIED but is broken out
					 * separately so that it may be modified in the future.
					 */
					rc = (*session->callback.notify)(dialog->session_id,
						dialog->server_response, -1L, dialog->input);
				}
			}
			else if (rc == 0) {
				/*
				 * if rc == 0, it means that the user connection closed.
				 */
				rc = ERROR;
				da_errno = DA_LOST_USER_CONNECTION;
				break;
			}
			break;

		default:
			break;
		}
	}
	else {
		switch (dialog->cmd_from_server) {
		/*
		 * The formats of the following commands are apparently undefined,
		 * so we ignore them.  They are noted for future expansion.
		 */
		case AGENT_ALIVE:
		case AGENT_STATUS:
		case NO_COMMAND:
			break;

		case SESSION_STATUS:
		case SESSION_CLEAR:
			/*
			 * DSS is requesting that this session be cleared.  We pass it
			 * along as an advisory only.
			 */
			session = get_session_by_tid(dialog->tid);
			if (session) {
				/*
				 * Call the notify function with a session id of -1
				 * to indicate that the TID must be used to look up the
				 * session.
				 */
				rc = (*session->callback.notify)(-1,
					dialog->cmd_from_server, dialog->tid, NULL);
			}
			break;

		default:
			break;
		}
	}
	if (dialog)
		free_dialog(dialog);
	return(rc);
}

/*
 * da_close_session()
 * 
 * Close an authentication session with the DSS.
 */
int
da_close_session(int session_id)
{
	session_t *session;
	
	da_errno = SUCCESS;
	/*
	 * Sanity check to make sure that the server is actually connected
	 */
	if (!agent_cfg || (agent_cfg->server_fd == -1)) {
		da_errno = DA_NO_CONFIGURE;
		return(ERROR);
	}

	/*
	 * Get the session from the list of active sessions.  Make sure
	 * that it is still active.
	 */

	if ((session = get_session_by_id(session_id)))
		free_session(session);

	return(SUCCESS);
}

/*
 * da_open_session()
 * 
 * Open an authentication session with the DSS.
 */
int
da_open_session(char *user, dialog_io_t *callback)
{
	session_t *session;
	dialog_t *dialog;
	int rc = SUCCESS;
	
	da_errno = SUCCESS;
	/*
	 * Sanity check to make sure that the server is actually connected
	 */
	if (!agent_cfg || (agent_cfg->server_fd == -1)) {
		da_errno = DA_NO_CONFIGURE;
		return(ERROR);
	}
	if (!callback->notify || !callback->write || !callback->read) {
		da_errno = DA_NO_CONFIGURE;
		return(ERROR);
	}

	session = alloc_session();
	if (!session) {
		if (callback->error)
			return((*callback->error)(-1, da_errno));
		else
			return(ERROR);
	}
		
	session->user = user;
	if (user)
		strncpy(session->userid, user, USER_ID_LEN);
	session->callback = *callback;

	/*
	 * Get a dialog.  alloc_dialog() will initialize the structure.
	 * We do not link the dialog onto the session because send_dialog()
	 * will free the dialog.
	 */
	dialog = alloc_dialog();
	if (!dialog) {
		if (session->callback.error)
			return((*session->callback.error)(session->session_id, da_errno));
		else
			return(ERROR);
	}

	if (user && *user != '\0') {
		dialog->cmd_from_agent = CHECK_ID;
		strncpy(dialog->input, user, INPUT_LEN);
		strncpy(session->userid, user, USER_ID_LEN);
	}
	else
		dialog->cmd_from_agent = SAW_DCD;

	/*
	 * Copy ALL the assigned session identifier values from the session 
	 * to the dialog.  Since this is the first time through, callslot
	 * and transaction are '-1'(assigned by server) and session id 
	 * should be unique.
	 */
	dialog->callslot = session->callslot;
	dialog->session_id = session->session_id;
	dialog->tid = session->transaction_id;

	if (send_dialog(agent_cfg->server_fd, dialog)) {
		if (session->callback.error)
			rc = (*session->callback.error)(session->session_id, da_errno);
		else
			rc = ERROR;
	}
	else 
		rc = session->session_id;
	return(rc);
}

/*
 * da_end_call()
 *
 * Send an end of call indication for a user, transaction id pair.
 * If the agent is not configured to support end of call, the request
 * is ignored.
 */
int
da_end_call(char *user, long tid)
{
	session_t *session;
	dialog_t *dialog;
	int rc = SUCCESS;
	
	da_errno = SUCCESS;
	/*
	 * Make sure that a real user name is passed in.
	 */
	if (!user) {
		da_errno = DA_INVALID_PARAMETER;
		return(ERROR);
	}
	/*
	 * Sanity check to make sure that the server is actually connected
	 */
	if (!agent_cfg || (agent_cfg->server_fd == -1)) {
		da_errno = DA_NO_CONFIGURE;
		return(ERROR);
	}

	/*
	 * If there is an existing session on the active list,
	 * use that session id and free the session.
	 */
	if ((session = get_session_by_tid(tid))) {
		/*
		 * Are we still connected to the same server?
		 * If not, don't send the End of Call indication because it
		 * will only confuse the DSS.  This should go away sometime...
		 */
		if (session->in_addr.s_addr !=
			((struct sockaddr_in *)&agent_cfg->sa)->sin_addr.s_addr) {
			free_session(session);
			return(SUCCESS);
		}
	}

	/*
	 * Do we support end of call?
	 */
	if (agent_cfg->da_cfg.flags & REPORT_END_OF_CALL) {
		/*
		 * Get a dialog.  alloc_dialog() will initialize the structure.
		 */
		if (!(dialog = alloc_dialog())) {
			if (session->callback.error)
				rc = (*session->callback.error)(session->session_id, da_errno);
			else
				rc = ERROR;
			return(rc);
		}
		dialog->cmd_from_agent = END_CALL;
		strncpy(dialog->input, user, USER_ID_LEN);

		/*
		 * Copy ALL the assigned session identifier values from the session 
		 * to the dialog.  Since this is the end of call we may not have a
		 * callslot assigned.  In this case, just use '-1' (assigned by
		 * server).
		 */
		dialog->tid = tid;

		/*
		 * If there is an existing session on the active list,
		 * use that session id and free the session.
		 */
		if (session) {
			dialog->callslot = session->callslot;
			dialog->session_id = session->session_id;
			free_session(session);
		}
		if (send_dialog(agent_cfg->server_fd, dialog)) {
			if (session->callback.error)
				rc = (*session->callback.error)(session->session_id, da_errno);
			else
				rc = ERROR;
		}
	}
	else if (session)
		free_session(session);

	return(rc);
}

/*============================================================================
 * da_authenticate()
 *
 * authenticate a user  
 * outputs:
 * 	return 0 on success, and user name  
 *	      -1 on error, 
 *		and the server command code on no authenticate.       
 *
 * Notes:
 * Removed calls to myread and mywrite
 *==========================================================================*/
int da_authenticate(char *user, dialog_io_t *callback, long *tid, void *env)
{
   session_t *session;
   dialog_t *dialog;
   int rc = 0;
	
   da_errno = SUCCESS;
   *tid = -1L;

   /*
    * Sanity check to make sure that the server is actually connected
    */
   if (!agent_cfg || (agent_cfg->server_fd == -1)) 
   {
      da_errno = DA_NO_CONFIGURE;
      return(ERROR);
   }
   if (!callback->write || !callback->read) 
   {
      da_errno = DA_NO_CONFIGURE;
      return(ERROR);
   }

   /*
    * Allocate a session and dialog for this transaction.  We do not
    * link the session and dialog because control remains local.
    */
   session = alloc_session();
   dialog = alloc_dialog();
   if (user && *user != '\0') 
   {
      dialog->cmd_from_agent = CHECK_ID;
      strncpy(dialog->input, user, INPUT_LEN);
      strncpy(session->userid, user, USER_ID_LEN);
   }
   else
      dialog->cmd_from_agent = SAW_DCD;

   /*
    * Copy ALL the assigned session identifier values from the session 
    * to the dialog.  Since this is the first time through, callslot
    * and transaction are '-1'(assigned by server) and session id 
    * should be unique.
    */
   dialog->callslot = session->callslot;
   dialog->session_id = session->session_id;
   dialog->tid = session->transaction_id;

   while (!dialog->call_complete && !rc) 
   {
      /*
       * Send the packet to the server and wait for a response
       */

      if ((rc = send_dialog(agent_cfg->server_fd, dialog)))
         break;

      /*
       * Wait for the server to respond.  The return code is non-zero
       * if either a timeout occurred or an error occurred.
       */
      rc = da_wait_server(NULL);
      if (rc <= 0) 
      {
         dialog = NULL;
	 break;
      }

      /*
       * Read the packet and parse it out.
       */
      if (!(dialog = process_dialog())) 
      {
         rc = ERROR;
	 break;
      }

      /*
       * Update the server callslot and transaction id fields
       * in the session block.
       */
      session->callslot = dialog->callslot;
      session->transaction_id = dialog->tid;

      switch (dialog->cmd_from_server) 
      {
	  case GET_INPUT:
	      /*
	       * write out the prompt string.  Write returns the length
	       * written out.  We currently do not check except for error
	       * conditions, but it's available later.
	       */
	      rc = (*callback->write)(dialog->session_id, dialog->prompt,
				strlen(dialog->prompt));
	      if (rc < 0)
	         break;
	      rc = 0;

	      /*
	       * get the user input
	       */
	      rc = (*callback->read)(dialog->session_id, dialog->input,
				INPUT_LEN, dialog->server_response, dialog->line_flags);
	      if (rc < 0)
	         break;
	      rc = 0; 
              if(strstr(dialog->prompt,"Enter ID:") )
              { 
                 strncpy(user,dialog->input,16);
              }
	      dialog->cmd_from_agent = USER_INPUT;
	      break;

	  case ACCESS_APPROVED:
			/*
			 * write out the prompt string
			 */           
/*
			rc = (*callback->write)(dialog->session_id,
				dialog->prompt, strlen(dialog->prompt));
			if (rc < 0)
				break;
*/
			rc = 0;
			*tid = dialog->tid;
			if (env)
				memcpy(env, dialog->input, INPUT_LEN);
			break;

		case ACCESS_DENIED:
			/*
			 * write out the prompt string
			 */
			rc = (*callback->write)(dialog->session_id,
				dialog->prompt, strlen(dialog->prompt));
			if (rc < 0)
				break;
			/*
			 * On a denial, set the transaction ID = -1;
			 */
			rc = dialog->server_response;
			*tid = -1L;
			if (env)
				memcpy(env, dialog->input, INPUT_LEN);
			break;

		case NO_COMMAND:
			/*
			 * We have two cases were the access is set.  It is possible
			 * to get a command granting or denying access, or there is no
			 * command and the response code carries the status of the access
			 * request.  This is pretty nonsensical, but must be dealt with.
			 * We only allow a limited set of the defined responses to occur
			 * here.
			 */
			/*
			 * write out the prompt string
			 */
			rc = (*callback->write)(dialog->session_id,
				dialog->prompt, strlen(dialog->prompt));
			if (rc < 0)
				break;

			if (dialog->server_response == ACCESS_APPROVED) {
				rc = SUCCESS;
				*tid = dialog->tid;
			}
			else if (dialog->server_response == ACCESS_DENIED) {
				/*
				 * On a denial, set the transaction ID = -1;
				 */
				rc = dialog->server_response;
				*tid = -1L;
			}
			else {
				/*
				 * This is the same as ACCESS_DENIED but is broken out
				 * separately so that it may be modified in the future.
				 */
				rc = dialog->server_response;
				*tid = -1L;
			}
			if (env)
				memcpy(env, dialog->input, INPUT_LEN);
			break;

		default:
			*tid = -1L;
			rc = ERROR;
			break;
		}
	}
	if (dialog)
		free_dialog(dialog);
	free_session(session);
	if (rc && callback->error)
		rc = (*callback->error)(-1, da_errno);

	return(rc);
}

/*
 * Connect/disconnect functions for defender agent
 *
 * Calling da_disconnect() explicitly will flush all current user sessions
 * Calling da_connect() will replace any existing connection and will 
 * leave the user sessions intact.
 */

void
da_disconnect(void)
{
    session_t *session;
    
	da_errno = SUCCESS;

	/*
	 * Sanity check to make sure that the configuration block
	 * has been allocated.
	 */
	if (agent_cfg) {
		for (session = agent_cfg->sessions; session; session = session->next)
			free_session(session);

	    /*
	     * Close the existing connection.
		 */
		if (agent_cfg->server_fd != -1)
#ifdef _WIN32
			CloseHandle((HANDLE)agent_cfg->server_fd);
#else
			close(agent_cfg->server_fd);
#endif

		free(agent_cfg);
		agent_cfg = NULL;
	}
}

/*
 * da_connect()
 * 
 * connect to a DSS.
 */
int
da_connect(struct sockaddr *sa, da_config_t *cfg)
{
	auth_t auth;
	dialog_t *dialog;
	unsigned long challenge;
    int rc = 0;
    
	da_errno = SUCCESS;

	if (!agent_cfg) {
		agent_cfg = malloc(sizeof(cfg_t));
		if (!agent_cfg) {
			da_errno = DA_ALLOC_FAILED;
			return(ERROR);
		}
		memset(agent_cfg, 0, sizeof(cfg_t));
		agent_cfg->server_fd = -1;
	}

    /*
     * If there is an existing connection, close it first,
	 * but do not close existing user session entries.
	 */
	if (agent_cfg->server_fd != -1)
#ifdef _WIN32
		CloseHandle((HANDLE)agent_cfg->server_fd);
#else
		close(agent_cfg->server_fd);
#endif

	/*
	 * Copy the user supplied configuration data and socket address
	 * into the agent configuration structure.
	 */
	agent_cfg->da_cfg = *cfg;
	agent_cfg->sa = *sa;

    /*
     * create a socket and connect to the server
     */
    agent_cfg->server_fd = socket(AF_INET, SOCK_STREAM, 0);
   	if (agent_cfg->server_fd < 0)
   		return(agent_cfg->server_fd);

   	rc = connect(agent_cfg->server_fd, sa, sizeof(struct sockaddr));
	if (rc < 0) {
#ifdef _WIN32
		CloseHandle((HANDLE)agent_cfg->server_fd);
#else
		close(agent_cfg->server_fd);
#endif

		agent_cfg->server_fd = -1;
		return(agent_cfg->server_fd);
	}

	if (!(dialog = alloc_dialog())) {
#ifdef _WIN32
		CloseHandle((HANDLE)agent_cfg->server_fd);
#else
		close(agent_cfg->server_fd);
#endif

		agent_cfg->server_fd = -1;
		return(agent_cfg->server_fd);
	}

	dialog->callslot = -1;
	dialog->session_id = -1;
	dialog->tid = -1L;
	dialog->cmd_from_agent = AGENT_ID;
	challenge = GetRandomLong();

	/*
	 * set up the authentication record.
	 */
	auth.state = DA_AS_ID;
	auth.level = AUTH_LEVEL;
	auth.challenge = challenge;
	auth.response = 0L;
	auth.session_key[0] = 0L;
	auth.session_key[1] = 0L;

	write_auth(dialog, &auth, agent_cfg->da_cfg.agent_id);

	if ((rc = send_dialog(agent_cfg->server_fd, dialog)))
		return(rc);
	/*
	 * Wait for the server to respond.  The return code is non-zero
	 * if either a timeout occurred or an error occurred.
	 */
	rc = da_wait_server(NULL);
	if (rc <= 0)
		return(ERROR);

	if(!(dialog = process_dialog()))
		return(ERROR);

	read_auth(dialog, &auth);

	/*
	 * Now we check ALL the appropriate fields for any sign of a
	 * rejection from the server.  The resulting huge conditional
	 * is really necessary.
	 * Note that at this point the auth.response field is the
	 * response to our challenge and the auth.challenge field is
	 * the challenge from the server that we must correctly encrypt.
	 */
	if (!(dialog->cmd_from_server == SERVER_CHALLENGE &&
		auth.state == DA_AS_CHALLENGE &&
		auth.response == snk_response(challenge))) {
#ifdef _WIN32
		CloseHandle((HANDLE)agent_cfg->server_fd);
#else
		close(agent_cfg->server_fd);
#endif

		agent_cfg->server_fd = -1;
		return(agent_cfg->server_fd);
	}

	dialog->cmd_from_agent = AGENT_RESPONSE;
	auth.state = DA_AS_RESPONSE;
	auth.response = snk_response(auth.challenge);

	/*
	 * Since the server has correctly answered our challenge, supply the
	 * agent configuration information.
	 */
	write_info(dialog);
	write_auth(dialog, &auth, agent_cfg->da_cfg.agent_id);

	if ((rc = send_dialog(agent_cfg->server_fd, dialog)))
		return(rc);
	/*
	 * Wait for the server to respond.  The return code is non-zero
	 * if either a timeout occurred or an error occurred.
	 */
	rc = da_wait_server(NULL);
	if (rc <= 0)
		return(ERROR);

	if (!(dialog = process_dialog()))
		return(ERROR);

	read_auth(dialog, &auth);
	
	if (!(dialog->cmd_from_server == SERVER_APPROVAL &&
		auth.state == DA_AS_APPROVAL)) {
#ifdef _WIN32
		CloseHandle((HANDLE)agent_cfg->server_fd);
#else
		close(agent_cfg->server_fd);
#endif
		agent_cfg->server_fd = -1;
		return(agent_cfg->server_fd);
	}

	/*
	 * We copy the options structure even though it is currently unused
	 */
	agent_cfg->options = *(options_t *)dialog->prompt;
	free_dialog(dialog);

	return(agent_cfg->server_fd);
}
